/*
 * 
 * Piotr Wjtowicz 05-10.2019
 * Przelacznik do zyrandola v2.5 / 2019
 * 
 * Kompilacja: XC8 v2.10
 *
 * Wprowadzone zmiany:
 * - pelnookresowe sterowane triakami
 * - attiny z fclk = 128kHz
 * - mniejszy prad bramek triakw
 * 
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include <stdint.h>

//#define TEST_BUILD
//#define FCLK_48 /* zegar 4,8MHz */
#define FCLK_128 /* zegar 128kHz */

#ifdef FCLK_48
// Fuses (E:FF, H:FD, L:71)
#define F_CPU   4800000UL
#define FCLK_PRES 4
#define TMR_PRES 8
#define TMR_OVF_MAX 256
#endif

#ifdef FCLK_128
// Fuses (E:FF, H:FF, L:7B)
#define F_CPU   128000UL
#define FCLK_PRES 1
#define TMR_PRES 1
#define TMR_OVF_MAX 218
#endif

#include <util/delay.h>

#define TMR1_COUNTER 12 /* divide factor for >timer1< long periods */ 
#define TMR_CLK_T ((float)(FCLK_PRES*TMR_PRES)/(F_CPU)) /*  */
#define TMR_OVF_T (TMR_OVF_MAX*TMR_CLK_T) /* periods of timer ovf events */
#define TMR1_OVF (TMR_OVF_T*TMR1_COUNTER) /* periods of timer1 ovf events */
#define MS2TMR(x) ((u8)((x)/(1000*TMR_OVF_T))) /* conversion from ms to timer0 events */
#define MS2TMR1(x) ((u8)((x)/(1000*TMR1_OVF))) /* conversion from miliseconds to timer1 ticks */
#define NZXT_MIN MS2TMR(25) /* minimal non zero cross events to detect power off */
#define ZXC_MIN 5 /* minimal required zero cross occurences */
#define MIN_DEBOUNCE_TIME_MS MS2TMR1(15) /* time for switch debouncing */
#define	EEADDR	0

/* -------------------------------------------------- */
typedef uint8_t u8;
typedef int8_t s8;
typedef uint16_t u16;
typedef int16_t s16;
typedef void (*statefun)(void);

typedef struct {
    statefun currentState;
    u8 zxc;
    u8 nzxt;
    u8 pwrt;
    u8 debt;    
    u8 lampt;
    u8 lampt_top;
} Pwrms;

volatile u8 tmr1;
volatile u8 lstate;
volatile Pwrms pwrms;

/* -------------------------------------------------- */
void pwr_off();
void pwr_on_i();
void pwr_on_c();
void pwr_poff_i();
void pwr_poff_c();
void pwr_poff_c_init();
void pwr_on_i_init();
u8 loadLampState();

/* -------------------------------------------------- */
static inline void init() {    
    PORTB |= (1 << PB3) | (1 << PB4); 			// set outputs as High
    DDRB |= 1 << PB3 | 1 << PB4; 				// pins as output  
    WDTCR |= (1<<WDE) | (1<<WDP2) | (1<<WDP1); 	// WDT 1s
    TCCR0A |= 1 << WGM01; 						// Set Timer0 CTC mode
	
#ifdef FCLK_48
    OCR0A = TMR_OVF_MAX-1;
    TCCR0B |= 1 << CS01; 						// Timer prescaler set to F_CPU/8
    CLKPR = 1 << CLKPCE;
    CLKPR = 1 << CLKPS1; 						// Clock prescaler 16x
#endif

#ifdef FCLK_128
    OCR0A = TMR_OVF_MAX-1;
    TCCR0B |= 1 << CS00; 						// Timer with no prescale
#endif
    
	TIMSK0 |= 1 << TOIE0 | 1 << OCIE0A; 		// timer0 overflow interrupt configuration
	MCUCR |= 1 << ISC01 | 1 << ISC00; 			// external int - trigger rising edge
	GIMSK |= 1 << INT0; 						// external int0 enabled
    lstate = loadLampState();
    pwrms.currentState = pwr_off;
	sei();
}

#ifdef TEST_BUILD
u8 lmp;
u8 lmpst;

inline void timer0blinkLedTest() {
    lmp++;
    if(lmp >= 25) {
        if(lmpst != 0) {
            lmpst = 0;
            setL1(0);

        } else {
            lmpst = 1;
            setL1(1);
        }
        lmp = 0;
    }
}
#endif

/* -------------------------------------------------- */
ISR(TIM0_COMPA_vect) {
    tmr1++;
    pwrms.nzxt++;    
    if(tmr1 >= TMR1_COUNTER) {
        pwrms.pwrt++;
        pwrms.lampt++;
		pwrms.debt++;
        tmr1 = 0;
    }
}


ISR(INT0_vect) {
	pwrms.nzxt = 0;
	
    if(pwrms.debt >= MIN_DEBOUNCE_TIME_MS) {
        pwrms.zxc++;
		pwrms.debt = MIN_DEBOUNCE_TIME_MS;
	} else {
		pwrms.debt = 0;
	}
}

/* -------------------------------------------------- */
/* Utils */
void setL(u8 p, u8 state) {
    PORTB = state ? PORTB & ~(1 << p) : PORTB | 1 << p;
}

void setL1(u8 state) {
    setL(PB3, state);
}

void setL2(u8 state) {
    setL(PB4, state);
}

void validateLampState(u8 *state) {
    if(*state < 1 || *state > 3) {
        *state = 3;
    }
}

u8 loadLampState() {
    u8 retVal;
	while(!eeprom_is_ready());
    retVal = eeprom_read_byte(EEADDR); 
    validateLampState(&retVal);
    return retVal;
}

void saveLampState(u8 *state) {
	validateLampState(state);
    if(loadLampState() == *state) {
        return;
    }
	while(!eeprom_is_ready());
    eeprom_write_byte(EEADDR, *state);
}

const u8 nextLampMap[] = { 0, 2, 3, 1 };

void nextLampState() {
    validateLampState(&lstate);
    lstate = nextLampMap[lstate];
}

void turnOnLamps() {
    if(lstate & 1) {
        setL1(1);
    } else {
        setL1(0);
    }
    if(lstate & 2) {
        setL2(1);
    } else {
        setL2(0);
    }
}

void turnOffLamps() {
    setL1(0);
    setL2(0);
}

/* -------------------------------------------------- */
/* State machine */
void pwr_off() {
    turnOffLamps();
    if(pwrms.zxc >= ZXC_MIN) {
        pwrms.lampt = 0;
        pwrms.lampt_top = MS2TMR1(100); /* delay light on after first touch the switch */
        pwrms.currentState = pwr_on_i;
    }
}

void pwr_on_i_init() {
    pwrms.nzxt = 0;
    pwrms.lampt = 0;
    pwrms.currentState = pwr_on_i;
}

void pwr_on_i() {
    if(pwrms.lampt > pwrms.lampt_top) {
        turnOnLamps();
        saveLampState(&lstate);
        pwrms.lampt = pwrms.lampt_top;
    }
    if(pwrms.nzxt > NZXT_MIN) {
        turnOffLamps();
        pwrms.nzxt = 0;
        pwrms.zxc = 0;
        pwrms.currentState = pwr_poff_i;
    }
}

void pwr_on_c() {	
    if(pwrms.pwrt > MS2TMR1(100)) { 	/* time since AC power on */
        pwrms.lampt_top = MS2TMR1(300); /* delay of light on after state change, [ms] */
        nextLampState();
        pwr_on_i_init();
    }
    if(pwrms.nzxt > NZXT_MIN) {
        turnOffLamps();
        pwr_poff_c_init();
    }    
}

void pwr_poff_c_init() {
    pwrms.zxc = 0;
    pwrms.currentState = pwr_poff_c;
}

void pwr_poff_i() {
    turnOffLamps();
    pwrms.pwrt = 0;
    if(pwrms.nzxt > MS2TMR(200)) {
        pwr_poff_c_init();
    }
    if(pwrms.zxc > ZXC_MIN) {
        pwrms.lampt_top = MS2TMR1(100);
        pwr_on_i_init();
    }
}

void pwr_poff_c() {
    turnOffLamps();
    pwrms.pwrt = 0;
    if(pwrms.zxc > 5) {
        pwrms.currentState = pwr_on_c;
    }
}

/* -------------------------------------------------- */
int main() {
    init();
    while(1) {
        pwrms.currentState();
        if(pwrms.currentState != pwr_poff_c) {
            asm("WDR");
        }
    }
	return 0;
}

